package aQute.library.remote;

import java.net.*;
import java.security.*;
import java.security.spec.*;
import java.text.*;
import java.util.*;

import aQute.lib.base64.*;
import aQute.lib.collections.*;
import aQute.lib.converter.*;
import aQute.lib.hex.*;
import aQute.service.library.*;

public class RemoteLibrary implements Library {
	static SimpleDateFormat							httpFormat			= new SimpleDateFormat(
																				"EEE, dd MMM yyyy HH:mm:ss zzz");
	final static TypeReference<Info>				singletonInfo		= new TypeReference<Info>() {};
	final static TypeReference<Program>				singletonProgram	= new TypeReference<Program>() {};
	final static TypeReference<Revision>			singletonRevision	= new TypeReference<Revision>() {};
	final static TypeReference<Iterable<Program>>	pluralProgram		= new TypeReference<Iterable<Program>>() {};
	final static TypeReference<Iterable<Revision>>	pluralRevision		= new TypeReference<Iterable<Revision>>() {};
	URLClient										urlc;
	PrivateKey										privateKey;
	PublicKey										publicKey;
	String											email;
	String											machine;

	public RemoteLibrary() {
		this(null);
	}

	public RemoteLibrary(String uri) {
		if (uri == null)
			uri = "http://repo.jpm4j.org/rest/";
		setUri(uri);
	}

	public void setUri(String uri) {
		urlc = new URLClient(uri) {

			protected void sign(HttpURLConnection con) throws Exception {
				if (privateKey == null)
					return;

				StringBuilder sb = new StringBuilder();
				sb.append("Basic ").append(email);

				if (machine != null)
					sb.append("!").append(machine);

				sb.append(":");

				String dateHeader = con.getRequestProperty("Date");
				if (dateHeader == null) {
					synchronized (httpFormat) {
						dateHeader = httpFormat.format(new Date());
					}
					con.setRequestProperty("Date", dateHeader);
				}

				Signature hmac = Signature.getInstance("SHA1withRSA");
				hmac.initSign(privateKey);
				hmac.update(dateHeader.getBytes()); // never UTF-8

				sb.append(Base64.encodeBase64(hmac.sign()));

				con.setRequestProperty("Authorization", sb.toString());
			}
		};
	}

	@Override
	public Program getProgram(String groupId, String artifactId) throws Exception {
		return urlc.get("program/" + groupId + "/" + artifactId, singletonProgram.getType(), null);
	}

	@Override
	public void update(Program program) throws Exception {
		urlc.put("library/program", program);
	}

	@Override
	public void redo(String where, String process) throws Exception {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean master(String coordinates) throws Exception {
		throw new UnsupportedOperationException();
	}

	@Override
	public Revision delete(String coordinates) {
		throw new UnsupportedOperationException();
	}

	class FindImpl<T> implements Find<T> {
		final MultiMap<String,String>		arguments	= new MultiMap<String,String>();
		final String						verb;
		final TypeReference<T>				singleton;
		final TypeReference<Iterable<T>>	plural;

		public FindImpl(String verb, TypeReference<T> singleton, TypeReference<Iterable<T>> plural) {
			this.verb = verb;
			this.singleton = singleton;
			this.plural = plural;
		}

		@Override
		public Find<T> skip(int n) throws Exception {
			arguments.add("skip", Integer.toString(n));
			return this;
		}

		@Override
		public Find<T> limit(int n) throws Exception {
			arguments.add("limit", Integer.toString(n));
			return this;
		}

		@Override
		public Find<T> ascending(String field) throws Exception {
			arguments.add("order", "+" + field);
			return this;
		}

		@Override
		public Find<T> descending(String field) throws Exception {
			arguments.add("order", "-" + field);
			return this;
		}

		@Override
		public Find<T> where(String where, Object... args) throws Exception {
			String formatted = String.format(where, args);
			arguments.add("where", formatted);
			return this;
		}

		@Override
		public T one() throws Exception {
			arguments.add("type", "one");
			return get(singleton);
		}

		@Override
		public T first() throws Exception {
			arguments.add("type", "first");
			return get(singleton);
		}

		@Override
		public int count() throws Exception {
			arguments.add("type", "count");
			return get(new TypeReference<Integer>() {});
		}

		@Override
		public Iterator<T> iterator() {
			try {
				arguments.add("type", "iterator");
				return get(plural).iterator();
			}
			catch (Exception e) {
				throw new RuntimeException(e);
			}
		}

		private <X> X get(TypeReference<X> ref) throws Exception {
			return urlc.get(verb, ref.getType(), arguments);
		}

		@Override
		public Find<T> bsn(String bsn) throws Exception {
			if (plural.equals(pluralProgram))
				return where("_id=%s", bsn);
			else
				return where("bsn=%s", bsn);
		}

		@Override
		public Find<T> version(String version) throws Exception {
			assert (!plural.equals(pluralProgram));
			return where("version=%s", version);
		}

		@Override
		public Find<T> baseline(String version) throws Exception {
			return where("baseline=%s", version);
		}

		@Override
		public Find<T> qualifier(String qualifier) throws Exception {
			return where("qualifier=%s", qualifier);
		}

		@Override
		public Find<T> from(long date) throws Exception {
			return where("insertDate>=%s", date);
		}

		@Override
		public Find<T> until(long date) throws Exception {
			return where("insertDate<%s", date);
		}

		@Override
		public Find<T> query(String query) throws Exception {
			arguments.add("query", query);
			return this;
		}

		@Override
		public Find<T> template(Revision rev) throws Exception {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public Find<T> capability(String ns, String key, Object value) throws Exception {
			arguments.add("capability", ns + "," + key + "," + value);
			return this;
		}

		@Override
		public boolean callback(Callback<T> visitor) throws Exception {
			throw new UnsupportedOperationException();
		}
	}

	@Override
	public Find<Program> findProgram() throws Exception {
		return new FindImpl<Program>("program", singletonProgram, pluralProgram);
	}

	@Override
	public Find<Revision> findRevision() throws Exception {
		return new FindImpl<Revision>("revision", singletonRevision, pluralRevision);
	}

	@Override
	public Info getInfo() throws Exception {
		FindImpl<Info> f = new FindImpl<Info>("library", singletonInfo, null);
		return f.one();
	}

	public void credentials(String email, String machine, byte[] publicKey, byte[] privateKey) throws Exception {
		this.email = email;
		this.machine = machine;

		if (publicKey != null && privateKey != null) {
			PKCS8EncodedKeySpec privateKeySpec = new PKCS8EncodedKeySpec(privateKey);
			X509EncodedKeySpec publicKeySpec = new X509EncodedKeySpec(publicKey);
			KeyFactory keyFactory = KeyFactory.getInstance("RSA");
			this.privateKey = keyFactory.generatePrivate(privateKeySpec);
			this.publicKey = keyFactory.generatePublic(publicKeySpec);
		}
	}

	@Override
	public Revision scan(URI uri) throws Exception {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Revision stage(Revision revision) throws Exception {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Revision stage(URI url) throws Exception {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Revision getRevision(byte[] sha) throws Exception {
		if (sha == null)
			throw new IllegalArgumentException("SHA must be specified");

		if (sha.length != 20)
			throw new IllegalArgumentException("Invalid SHA must be 20 bytes");

		return urlc.get("revision/sha/" + Hex.toHexString(sha) + "//0.0.0", singletonRevision.getType(), null);
	}

	@Override
	public Revision getRevision(String groupId, String artifactId, String classifier, String version) throws Exception {
		if (classifier == null || classifier.trim().isEmpty())
			classifier = "";

		return urlc.get("revision/" + groupId + "/" + artifactId + "/" + classifier + "/" + version, singletonRevision
				.getType(), null);
	}

	@Override
	public Category getCategory(String name) throws Exception {
		throw new UnsupportedOperationException();
	}

	@Override
	public Iterable< ? extends Category> getCategories(String where) throws Exception {
		throw new UnsupportedOperationException();
	}

	@Override
	public void update(Category c) throws Exception {
		throw new UnsupportedOperationException();
	}

	@Override
	public Revision rescan(String coordinates) throws Exception {
		throw new UnsupportedOperationException();
	}

	@Override
	public Revision getRevision(String coordinates) throws Exception {
		return urlc.get("revision/" + coordinates, singletonRevision.getType(), null);
	}

	@Override
	public Iterable<Revision> getRevisions(String coordinates) throws Exception {
		return urlc.get("revisions/" + coordinates, pluralRevision.getType(), null);
	}

	@Override
	public void scan(ScanRequest rq) throws Exception {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean update(Program p, String... field) throws Exception {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean update(Revision r, String... field) throws Exception {
		throw new UnsupportedOperationException();
	}

	@Override
	public Iterable< ? extends Program> getPrograms(String coordinates) throws Exception {
		return urlc.get("programs/" + coordinates, pluralProgram.getType(), null);
	}
}
