package aQute.maven;

import java.io.*;

public class MavenVersionRange {
	final MavenVersionRange	and;
	final int				start;
	final int				end;
	final MavenVersion		low;
	final MavenVersion		high;

	static class Parser {
		int		current;
		Reader	rdr;

		public Parser(StringReader rdr) throws Exception {
			this.rdr = rdr;
			readAndSkip();
		}

		int skipWs() throws IOException {
			while (current >= 0 && Character.isWhitespace(current))
				current = rdr.read();

			return current;
		}

		public MavenVersionRange parseRange() throws Exception {
			skipWs();
			switch (current) {
				case '[' :
				case '(' :

					int start = current;
					readAndSkip();
					MavenVersion low = parseVersion();
					MavenVersion high = null;
					;
					if (current == ',') {
						read();
						skipWs();
						high = parseVersion();
					}
					skipWs();
					int end = current;
					if (end != ']' && end != ')')
						throw new IllegalArgumentException("Not a valid maven range");

					read();
					skipWs();
					MavenVersionRange next = null;
					if (current == ',') {
						read();
						skipWs();
						next = parseRange();
					}
					return new MavenVersionRange(start, low, high, end, next);

				default :
					return new MavenVersionRange('[', parseVersion(), null, ')', null);
			}
		}

		private int readAndSkip() throws Exception {
			read();
			skipWs();
			return current;
		}

		private void read() throws IOException {
			if (current >= 0)
				current = rdr.read();
		}

		private MavenVersion parseVersion() throws IOException {
			int major = parseInt();
			int minor = 0;
			int increment = 0;
			String qualifier = "";

			if (current == '.') {
				read();
				minor = parseInt();
				if (current == '.') {
					read();
					increment = parseInt();
				}
			}
			if (current == '-') {
				read();
				qualifier = parseQualifier();
			}
			return new MavenVersion(major, minor, increment, qualifier);
		}

		private String parseQualifier() throws IOException {
			StringBuilder sb = new StringBuilder();

			while ((current >= '0' && current <= '9') || (current >= 'a' && current <= 'z')
					|| (current >= 'A' && current <= 'Z') || current == '-' || current == '_' || current == '.') {
				sb.append((char) current);
				read();
			}

			return sb.toString();
		}

		private int parseInt() throws IOException {
			int n = 0;
			while (current >= '0' && current <= '9') {
				n = n * 10 + current - '0';
				read();
			}

			return n;
		}
	}

	public MavenVersionRange(int start, MavenVersion low, MavenVersion high, int end, MavenVersionRange next) {
		this.start = start;
		this.low = low;
		this.high = high;
		this.end = end;
		this.and = next;
	}

	public static MavenVersionRange parse(String v) throws Exception {
		Parser p = new Parser(new StringReader(v));
		try {
			return p.parseRange();
		}
		catch (Exception e) {
			return null;
		}
	}

	public String toString() {
		StringBuilder sb = new StringBuilder();
		toString(sb);
		return sb.toString();
	}

	void toString(StringBuilder sb) {
		if (start == '[' && low != null && high == null && end == ')') {
			low.toString(sb);
			return;
		} else if (start == '[' && low != null && high == null && end == ']') {
			sb.append('[').append(low).append(']');
			return;
		}

		sb.append((char) start);
		if (low != null)
			sb.append(low);
		sb.append(",");
		if (high != null)
			sb.append(high);

		if (and != null) {
			sb.append(",");
			and.toString(sb);
		}
	}

	public String toFilter() {
		StringBuilder sb = new StringBuilder();
		toFilter(sb);
		return sb.toString();
	}

	public void toFilter(StringBuilder sb) {
		if (and == null) {
			localToFilter(sb);
		} else {
			sb.append("(&");
			localToFilter(sb);
			sb.append(")");
		}
	}

	private void localToFilter(StringBuilder sb) {
		sb.append("(");
		int n = 0;
		if (start == '[')
			n += 1;

		if (low != null)
			n += 2;

		if (high != null)
			n += 4;

		if (start == ']')
			n += 8;

		switch (n) {
			case 0 : // (,)
			case 1 : // [,)
			case 8 : // (,]
			case 9 : // [,]
				sb.append("v=*");
				break;

			case 10 : // (x,]
			case 2 : // (x,)
				sb.append("!(v<=").append(low).append(")");
				break;

			case 11 : // [x,]
			case 3 : // [x,)
				sb.append("v>=").append(low);
				break;

			case 4 : // (,x)
			case 5 : // [,x)
				sb.append("!(v>=").append(high).append(")");
				break;

			case 12 : // (,x]
			case 13 : // [,x]
				sb.append("v<=").append(high);
				break;

			case 6 : // (x,y)
				sb.append("&(!(v<=").append(low).append("))(!(v<").append(high).append("))");
				break;

			case 7 : // [x,y)
				sb.append("&(v>=").append(low).append(")(!(v<").append(high).append("))");
				break;

			case 14 : // (x,y]
				sb.append("&(!(v<=").append(low).append("))(v<=").append(high).append(")");
				break;

			case 15 : // [x,y]
				sb.append("&(v>=").append(low).append(")(v<=").append(high).append(")");
				break;

		}

		sb.append(")");
		if (and != null)
			and.toFilter(sb);
	}
}
