package aQute.maven;

import java.net.*;
import java.util.*;

import aQute.struct.*;

/**
 * TODO finish this to support the complete pom. This is not a normal project
 * pom but only holds the values for poms in the repository. I.e. this lacks
 * lots of build stuff.
 */
public class Pom extends BasicPom {
	/**
	 * The packaging type required to be pom for parent and aggregation
	 * (multi-module) projects. These types define the goals bound to a set of
	 * lifecycle stages. For example, if packaging is jar, then the package
	 * phase will execute the jar:jar goal. If the packaging is pom, the goal
	 * executed will be site:attach-descriptor. Now we may add values to the
	 * parent POM, which will be inherited by its children. The elements in the
	 * parent POM that are inherited by its children are: dependencies
	 * developers and contributors plugin lists reports lists plugin executions
	 * with matching ids plugin configuration
	 */

	public static class PomRef extends struct {
		public String	groupId;
		public String	artifactId;
		public String	version;
		public String	relativePath;
		public String	message;
	}

	/**
	 * exclusions: Exclusions contain one or more exclusion elements, each
	 * containing a groupId and artifactId denoting a dependency to exclude.
	 * Unlike optional, which may or may not be installed and used, exclusions
	 * actively remove themselves from the dependency tree.
	 */
	public static class Exclusion extends struct {
		public String	groupId;
		public String	artifactId;
	}

	/**
	 * Not sure where this is defined ...
	 */
	public static class DependencyManagement extends struct {
		public List<Dependency>	dependencies	= list();
	}

	/**
	 * One powerful aspect of Maven is in its handling of project relationships;
	 * that includes dependencies (and transitive dependencies), inheritance,
	 * and aggregation (multi-module projects). Dependency management has a long
	 * tradition of being a complicated mess for anything but the most trivial
	 * of projects. "Jarmageddon" quickly ensues as the dependency tree becomes
	 * large and complicated. "Jar Hell" follows, where versions of dependencies
	 * on one system are not equivalent to versions as those developed with,
	 * either by the wrong version given, or conflicting versions between
	 * similarly named jars. Maven solves both problems through a common local
	 * repository from which to link projects correctly, versions and all. </p>
	 * The cornerstone of the POM is its dependency list. Most every project
	 * depends upon others to build and run correctly, and if all Maven does for
	 * you is manage this list for you, you have gained a lot. Maven downloads
	 * and links the dependencies for you on compilation and other goals that
	 * require them. As an added bonus, Maven brings in the dependencies of
	 * those dependencies (transitive dependencies), allowing your list to focus
	 * solely on the dependencies your project requires.
	 */
	public static class Dependency extends struct {
		/**
		 * This element refers to the classpath of the task at hand (compiling
		 * and runtime, testing, etc.) as well as how to limit the transitivity
		 * of a dependency. There are five scopes available:
		 */
		public String			scope;
		/**
		 * Corresponds to the dependant artifact's packaging type. This defaults
		 * to jar. While it usually represents the extension on the filename of
		 * the dependency, that is not always the case. A type can be mapped to
		 * a different extension and a classifier. The type often corresponds to
		 * the packaging used, though this is also not always the case. Some
		 * examples are jar, ejb-client and test-jar. New types can be defined
		 * by plugins that set extensions to true, so this is not a complete
		 * list.
		 */
		public String			type;
		/**
		 * Marks optional a dependency when this project itself is a dependency.
		 * Confused? For example, imagine a project A that depends upon project
		 * B to compile a portion of code that may not be used at runtime, then
		 * we may have no need for project B for all project. So if project X
		 * adds project A as its own dependency, then Maven will not need to
		 * install project B at all. Symbolically, if => represents a required
		 * dependency, and --> represents optional, although A=>B may be the
		 * case when building A X=>A-->B would be the case when building X.
		 */
		public boolean			optional;
		/**
		 * These elements are self-explanatory, and you will see them often.
		 * This trinity represents the coordinate of a specific project in time,
		 * demarcating it as a dependency of this project. You may be thinking:
		 * "This means that my project can only depend upon Maven artifacts!"
		 * The answer is, "Of course, but that's a good thing." This forces you
		 * to depend solely on dependencies that Maven can manage. There are
		 * times, unfortunately, when a project cannot be downloaded from the
		 * central Maven repository. For example, a project may depend upon a
		 * jar that has a closed-source license which prevents it from being in
		 * a central repository. There are three methods for dealing with this
		 * scenario. Install the dependency locally using the install plugin.
		 * The method is the simplest recommended method. For example: mvn
		 * install:install-file -Dfile=non-maven-proj.jar -DgroupId=some.group
		 * -DartifactId=non-maven-proj -Dversion=1 -Dpackaging=jar Notice that
		 * an address is still required, only this time you use the command line
		 * and the install plugin will create a POM for you with the given
		 * address. Create your own repository and deploy it there. This is a
		 * favorite method for companies with an intranet and need to be able to
		 * keep everyone in synch. There is a Maven goal called
		 * deploy:deploy-file which is similar to the install:install-file goal
		 * (read the plugin's goal page for more information). Set the
		 * dependency scope to system and define a systemPath. This is not
		 * recommended, however, but leads us to explaining the following
		 * elements:
		 */
		public String			groupId;
		public String			artifactId;
		public String			version;
		/**
		 * The classifier allows to distinguish artifacts that were built from
		 * the same POM but differ in their content. It is some optional and
		 * arbitrary string that - if present - is appended to the artifact name
		 * just after the version number. As a motivation for this element,
		 * consider for example a project that offers an artifact targeting JRE
		 * 1.5 but at the same time also an artifact that still supports JRE
		 * 1.4. The first artifact could be equipped with the classifier jdk15
		 * and the second one with jdk14 such that clients can choose which one
		 * to use. Another common use case for classifiers is the need to attach
		 * secondary artifacts to the project's main artifact. If you browse the
		 * Maven central repository, you will notice that the classifiers
		 * sources and javadoc are used to deploy the project source code and
		 * API docs along with the packaged class files.
		 */
		public String			classifier;

		/**
		 * Exclusions explicitly tell Maven that you don't want to include the
		 * specified project that is a dependency of this dependency (in other
		 * words, its transitive dependency). For example, the maven-embedder
		 * requires maven-core, and we do not wish to use it or its
		 * dependencies, then we would add it as an exclusion.
		 */
		public List<Exclusion>	exclusions	= list();
		/**
		 * is used only if the the dependency scope is system. Otherwise, the
		 * build will fail if this element is set. The path must be absolute, so
		 * it is recommended to use a property to specify the machine-specific
		 * path (more on properties below), such as ${java.home}/lib. Since it
		 * is assumed that system scope dependencies are installed a priori,
		 * Maven will not check the repositories for the project, but instead
		 * checks to ensure that the file exists. If not, Maven will fail the
		 * build and suggest that you download and install it manually.
		 */
		public String			systemPath;
	}

	public static class Release extends struct {

		public boolean	enabled;
		public String	updatePolicy;
		public String	checksumPolicy;
		public String	layout;

	}

	public static class Repository extends struct {
		public Release	releases;
		public Release	snapshots;
		public String	id;
		public String	name;
		public String	url;
	}

	/**
	 * Distribution management acts precisely as it sounds: it manages the
	 * distribution of the artifact and supporting files generated throughout
	 * the build process. Starting with the last elements first:
	 * <p/>
	 */
	public static class DistributionManagement extends struct {
		public List<Repository>	repositories	= list();
		/**
		 * downloadUrl: is the url of the repository from whence another POM may
		 * point to in order to grab this POM's artifact. In the simplest terms,
		 * we told the POM how to upload it (through repository/url), but from
		 * where can the public download it? This element answers that question.
		 */
		public String			downloadUrl;
		/**
		 * Warning! Like a baby bird in a nest, the status should never be
		 * touched by human hands! The reason for this is that Maven will set
		 * the status of the project when it is transported out to the
		 * repository. Its valid types are as follows.
		 * <ul>
		 * <li>none: No special status. This is the default for a POM.</li>
		 * <li>converted: The manager of the repository converted this POM from
		 * an earlier version to Maven 2.</li>
		 * <li>partner: This could just as easily have been called synched. This
		 * means that this artifact has been synched with a partner repository.</li>
		 * <li>deployed: By far the most common status, meaning that this
		 * artifact was deployed from a Maven 2 or 3 instance. This is what you
		 * get when you manually deploy using the command-line deploy phase.</li>
		 * <li>verified: This project has been verified, and should be
		 * considered finalized.</li>
		 * </ul>
		 */
		public String			status;
		/**
		 * Projects are not static; they are living things (or dying things, as
		 * the case may be). A common thing that happens as projects grow, is
		 * that they are forced to move to more suitable quarters. For example,
		 * when your next wildly successful open source project moves under the
		 * Apache umbrella, it would be good to give your users as heads-up that
		 * the project is being renamed to org.apache:my-project:1.0. Besides
		 * specifying the new address, it is also good form to provide a message
		 * explaining why.
		 */
		public PomRef			relocation;
	}

	public PomRef					parent;
	/**
	 * Besides inheriting certain top-level elements, parents have elements to
	 * configure values for child POMs and transitive dependencies. One of those
	 * elements is dependencyManagement. dependencyManagement: is used by POMs
	 * to help manage dependency information across all of its children. If the
	 * my-parent project uses dependencyManagement to define a dependency on
	 * junit:junit:4.0, then POMs inheriting from this one can set their
	 * dependency giving the groupId=junit and artifactId=junit only, then Maven
	 * will fill in the version set by the parent. The benefits of this method
	 * are obvious. Dependency details can be set in one central location, which
	 * will propagate to all inheriting POMs. In addition, the version and scope
	 * of artifacts which are incorporated from transitive dependencies may also
	 * be controlled by specifying them in a dependency management section.
	 */
	public DependencyManagement		dependencyManagement;
	/**
	 * Not sure what the purpose of dependency management and top level
	 * dependencies is
	 */
	public List<Dependency>			dependencies	= list();

	/**
	 * Maintains errors and warnings when parsing the POM
	 */
	public List<String>				errors			= new ArrayList<String>();
	/**
	 * Maintains errors and warnings when parsing the POM
	 */
	public List<String>				warnings		= new ArrayList<String>();

	/**
	 * Distribution Management
	 */
	public DistributionManagement	distributionManagement;

	/**
	 * Guess on when this pom was created, usually last modified time of the web
	 * server
	 */
	public long						created;

	/**
	 * The repository URI
	 */

	public URI						repository;
	public URI						url;
}
